from tko.settings.settings import Settings
from tko.game.game import Game
from tko.settings.repository import Repository
from tko.play.play import Play
from tko.util.logger import Logger, LogAction
from typing import Dict
from tko.util.text import Text
import os

class CmdPlay:
    def __init__(self, settings: Settings):
        self.settings = settings
        self.rep: Repository | None = None
        self.alias = ""
        self.folder = ""

    def set_alias(self, rep_alias: str):
        if self.settings.has_alias_folder(rep_alias):
            folder = self.settings.get_alias_folder(rep_alias)
            rep = Repository(folder)
            if rep.has_local_config_file():
                rep.load_data_from_config_file().load_game()
                print(Text("Abrindo repositório de {g} em {g}", rep_alias, folder))
                self.rep = rep
                self.alias = rep_alias
                return self
            else:
                print(f"Repositório {rep_alias} não encontrado em {folder}")
                self.settings.del_alias_folder(rep_alias)
                # vai para o próximo bloco

        if rep_alias in self.settings.dict_alias_remote:
            print(Text("Repositório remoto {g} encontrado", rep_alias))
            folder = CmdPlay.choose_destiny(rep_alias)
            os.makedirs(folder, exist_ok=True)
            rep = Repository(folder)
            rep.set_remote_source(self.settings.get_alias_remote(rep_alias))
            rep.save_data_to_config_file()
            rep.load_data_from_config_file()
            rep.load_game()
            self.settings.set_alias_folder(rep_alias, folder)
            self.rep = rep
            self.alias = rep_alias
            return self
        
        raise Warning(Text("Repositório remoto {r} não encontrado", rep_alias))

    def load_folder(self, folder: str):
        self.folder = folder
        
        self.rep = Repository(folder)
        if not self.rep.has_local_config_file():
            print("Arquivo de configuração do repositório não encontrado")
            print("Você deseja criar um repositorio na pasta {}? ".format(folder), end="")
            print("s/n: ", end="")
            op = input()
            if op == "s":
                self.rep.create_empty_config_file(point_to_local_readme=True)
                self.settings.set_alias_folder(self.alias, folder)
                print("Repositório criado com sucesso")
            else:
                raise Warning("Repositório não encontrado")
        self.rep.load_data_from_config_file().load_game()
        return self


    def execute(self):
        if self.rep is None:
            raise Warning("Repositório não encontrado")
        
        logger: Logger = Logger.get_instance()
        logger.set_history_file(self.rep.get_history_file())
        logger.record_other_event(LogAction.OPEN)
        # target = ""
        # if self.folder != "":
        #     target = self.folder
        # else:
        #     target = self.alias
        play = Play(self.settings, self.rep)
        logger.set_daily(self.rep.get_daily_file(), play.game.tasks)
        play.play()
        logger.record_other_event(LogAction.QUIT)


    @staticmethod
    def choose_destiny(rep_alias):
        print(Text().add("Repositório local ainda não foi criado, onde você deseja criá-lo?"))
        here_cwd = os.path.join(os.getcwd(), rep_alias)
        qxcode = os.path.join(os.path.expanduser("~"), "qxcode", rep_alias)

        while True:
            print(Text().addf("r", "1").add(" - ").add(here_cwd))
            print(Text().addf("r", "2").add(" - ").add(qxcode))
            print(Text().addf("r", "3").add(" - ").add("Outra pasta"))
            print(Text().add("Default ").addf("r", "1").add(": "), end="")
            op = input()
            if op == "":
                op = "1"
            if op == "1":
                home_qxcode = here_cwd
                break
            if op == "2":
                home_qxcode = qxcode
                break
            if op == "3":
                print(Text().addf("y", "Navegue até o diretório desejado e execute o tko novamente."))
                exit(1)

        if not os.path.exists(home_qxcode):
            os.makedirs(home_qxcode)
        return home_qxcode


from typing import Any, Dict, List
import os
import json
import urllib
from ..util.remote_url import RemoteUrl, Absolute
from tko.game.game import Game
from tko.game.task import Task

languages_avaliable = ["c", "cpp", "py", "ts", "js", "java", "go"]

class Repository:
    CFG_FILE = "repository.json"
    LOG_FILE = "history.csv"
    DAILY_FILE = "daily.json"
    INDEX_FILE = "Readme.md"
    REMOTE_TASK_FOLDER = "remote"
    CONFIG_FOLDER = ".tko"

    def __init__(self, folder: str):
        self.folder: str = folder
        self.data: Dict[str, Any] = {}
        self.game = Game()

    @staticmethod
    def rec_search_for_repo(folder: str) -> str:
        if os.path.exists(os.path.join(folder, Repository.CONFIG_FOLDER, Repository.CFG_FILE)):
            return folder
        if folder == "/":
            return ""
        return Repository.rec_search_for_repo(os.path.abspath(os.path.join(folder, "..")))

    def get_remote_task_folder(self, label: str) -> str:
        return os.path.abspath(os.path.join(self.folder, Repository.REMOTE_TASK_FOLDER, label))

    def is_local_dir(self, path):
        rep_dir = self.get_rep_dir()
        path = os.path.abspath(path)
        return os.path.commonpath([rep_dir, path]) == rep_dir

    @staticmethod
    def is_web_link(link: str) -> bool:
        return link.startswith("http:") or link.startswith("https:")

    def set_task_folder(self, task: Task):
        if not task.is_downloadable():
            if not self.is_web_link(task.link):
                task_dir = os.path.abspath(os.path.dirname(task.link))
                task.set_folder(task_dir)
        else: # downloadable task
            task_dir = self.get_remote_task_folder(task.key)
            task.set_folder(task_dir)

    def load_game(self):
        self.game.parse_file(self.load_index_or_cache())
        for task in self.game.tasks.values():
            self.set_task_folder(task)
        return self

    def get_repository_file(self) -> str:
        return os.path.abspath(os.path.join(self.folder, Repository.CONFIG_FOLDER, Repository.CFG_FILE))
    
    def get_history_file(self) -> str:
        return os.path.abspath(os.path.join(self.folder, Repository.CONFIG_FOLDER, Repository.LOG_FILE))
    
    def get_daily_file(self) -> str:
        return os.path.abspath(os.path.join(self.folder, Repository.CONFIG_FOLDER, Repository.DAILY_FILE))
    
    def get_default_readme_path(self) -> str:
        return os.path.abspath(os.path.join(self.folder, Repository.INDEX_FILE))

    def get_rep_dir(self) -> str:
        return os.path.abspath(self.folder)

    def load_index_or_cache(self) -> str:
        file: str = self.get_local_file()
        remote: str = self.get_remote_source()
        # arquivo existe e é local
        if file != "" and os.path.exists(file) and remote == "":
            return file
        
        if not remote.startswith("http:)") and not remote.startswith("https:"):
            return remote

        # arquivo não existe e é remoto
        if remote != "" and (file == "" or not os.path.exists(file)):
            cache_file = self.get_default_readme_path()
            os.makedirs(self.folder, exist_ok=True)
            ru = RemoteUrl(remote)
            try:
                ru.download_absolute_to(cache_file)
            except urllib.error.URLError:
                print("Não foi possível baixar o arquivo do repositório")
                if os.path.exists(cache_file):
                    print("Usando arquivo do cache")
                else:
                    raise Warning("fail: Arquivo do cache não encontrado")
            return cache_file

        raise Warning("fail: arquivo não encontrado ou configurações inválidas para o repositório")

    __version = "version"
    __remote_url = "remote"  #remote url
    __local_file = "local"  #local file
    __expanded = "expanded"
    __new_items = "new_items"
    __tasks = "tasks"
    __flags = "flags"
    __lang = "lang"
    __selected = "index"

    defaults = {
        __version: "0.0.0",
        __remote_url: "",
        __local_file: "",
        __expanded: [],
        __tasks: {},
        __flags: {},
        __new_items: [],
        __lang: "",
        __selected: ""
    }

    def get_version(self) -> str:
        return self.__get(Repository.__version)
    
    def set_version(self, value: str):
        self.__set(Repository.__version, value)
        return self

    def get_local_file(self) -> str:
        return self.__get(Repository.__local_file)
    
    def set_local_file(self, value: str):
        self.__set(Repository.__local_file, value)
        return self

    def get_remote_source(self) -> str:
        return self.__get(Repository.__remote_url)
    
    def set_remote_source(self, value: str):
        self.__set(Repository.__remote_url, value)
        return self

    def get_selected(self) -> str:
        return self.__get(Repository.__selected)

    def get_expanded(self) -> List[str]:
        return self.__get(Repository.__expanded)

    def get_new_items(self) -> List[str]:
        return self.__get(Repository.__new_items)
    
    def get_tasks(self) -> Dict[str, Any]:
        return self.__get(Repository.__tasks)
    
    def get_flags(self) -> Dict[str, Any]:
        return self.__get(Repository.__flags)
    
    def get_lang(self) -> str:
        return self.__get(Repository.__lang)

    def set_expanded(self, value: List[str]):
        return self.__set(Repository.__expanded, value)
    
    def set_new_items(self, value: List[str]):
        return self.__set(Repository.__new_items, value)
    
    def set_tasks(self, value: Dict[str, str]):
        self.__set(Repository.__tasks, value)
        return self
    
    def set_flags(self, value: Dict[str, Any]):
        self.__set(Repository.__flags, value)
        return self
    
    def set_lang(self, value: str):
        self.__set(Repository.__lang, value)
        return self
    
    def set_selected(self, value: str):
        self.__set(Repository.__selected, value)
        return self

    def __get(self, key: str) -> Any:
        if key not in self.defaults:
            raise ValueError(f"Key {key} not found in RepSettings")
        value = self.data.get(key, Repository.defaults[key])
        if type(value) != type(Repository.defaults[key]):
            return Repository.defaults[key]
        return value

    def __set(self, key: str, value: Any):
        self.data[key] = value
        return self

    def create_empty_config_file(self, point_to_local_readme: bool = False):
        if point_to_local_readme:
            self.set_local_file(self.get_default_readme_path())
        self.save_data_to_config_file()
        return self

    def has_local_config_file(self) -> bool:
        return os.path.exists(self.get_repository_file())

    def load_data_from_config_file(self):
        with open(self.get_repository_file(), encoding="utf-8") as f:
            self.data = json.load(f)
        return self

    def save_data_to_config_file(self):
        json_file = self.get_repository_file()
        if not os.path.exists(os.path.dirname(json_file)):
            os.makedirs(os.path.dirname(json_file))
        # filter keys that are not in defaults
        for key in list(self.data.keys()):
            if key not in Repository.defaults:
                del self.data[key]

        for key in Repository.defaults:
            if key not in self.data:
                self.data[key] = Repository.defaults[key]

        with open(json_file, "w", encoding="utf-8") as f:
            f.write(json.dumps(self.data, indent=4))
        return self

    def __str__(self) -> str:
        return ( f"data: {self.data}\n" )



from tko.down.drafts import Drafts


import os
import urllib.error
import urllib.request
import tempfile
from tko.util.remote_url import RemoteUrl
from typing import Callable, Tuple


class DownProblem:
    fnprint: Callable[[str], None] = print

    @staticmethod
    def __create_file(content, path, label=""):
        with open(path, "w", encoding="utf-8") as f:
            f.write(content)
        DownProblem.fnprint(path + " " + label)

    @staticmethod
    def unpack_json(loaded, destiny, lang: str):
        # extracting all files to folder
        for entry in loaded["upload"]:
            if entry["name"] == "vpl_evaluate.cases":
                DownProblem.__compare_and_save(entry["contents"], os.path.join(destiny, "cases.tio"))
            else:
                DownProblem.__compare_and_save(entry["contents"], os.path.join(destiny, entry["name"]))

        for entry in loaded["required"]:
            DownProblem.__compare_and_save(entry["contents"], os.path.join(destiny, entry["name"]))
        
        for entry in loaded["keep"]:
            DownProblem.__compare_and_save(entry["contents"], os.path.join(destiny, entry["name"]))

        if "draft" in loaded:
            if lang in loaded["draft"]:
                for file in loaded["draft"][lang]:
                    path = os.path.join(destiny, file["name"])
                    DownProblem.__create_file(file["contents"], path, "(Rascunho)")

    @staticmethod
    def __compare_and_save(content: str, path: str):
        if not os.path.exists(path):
            with open(path, "w", encoding="utf-8") as f:
                f.write(content.encode("utf-8").decode("utf-8"))
            DownProblem.fnprint(path + " (Novo)")
        else:
            if open(path, encoding="utf-8").read() != content:
                DownProblem.fnprint(path + " (Atualizado)")
                with open(path, "w", encoding="utf-8") as f:
                    f.write(content)
            else:
                DownProblem.fnprint(path + " (Inalterado)")

    @staticmethod
    def down_readme(readme_path: str,  remote_url: RemoteUrl):
        temp_file = tempfile.mktemp()
        remote_url.download_absolute_to(temp_file)
        content = open(temp_file, encoding="utf-8").read()
        DownProblem.__compare_and_save(content, readme_path)
    
    @staticmethod
    def create_problem_folder(destiny: str):
        if not os.path.exists(destiny):
            os.makedirs(destiny, exist_ok=True)
        else:
            DownProblem.fnprint("Pasta do problema "+ destiny + " encontrada, juntando conteúdo.")

    @staticmethod
    def check_draft_existence(loaded_json, destiny: str, language: str, cache_url: str) -> bool:
        if len(loaded_json["required"]) > 0:  # you already have the students file
            for entry in loaded_json["required"]:
                if entry["name"].endswith("." + language):
                    return True

        if "draft" in loaded_json and language in loaded_json["draft"]:
            return True
        try:
            draft_path = os.path.join(destiny, "draft." + language)
            urllib.request.urlretrieve(cache_url + "draft." + language, draft_path)
            DownProblem.fnprint(draft_path + " (Rascunho)")
            return True
        except urllib.error.HTTPError:  # draft not found
            return False

    @staticmethod
    def create_default_draft(destiny: str, language: str):
        filename = "draft."
        draft_path = os.path.join(destiny, filename + language)
        if not os.path.exists(draft_path):
            with open(draft_path, "w", encoding="utf-8") as f:
                if language in Drafts.drafts:
                    f.write(Drafts.drafts[language])
                else:
                    f.write("")
            DownProblem.fnprint(draft_path + " (Vazio)")
        else:
            DownProblem.fnprint(draft_path + " (Não sobrescrito)")


import json
import os

class TaskData:
    def __init__(self, key: str = "", progress: int = 0, seft_grade: int = 0):
        self.key = key
        self.progress = progress
        self.seft_grade = seft_grade

    def decode(self, serial: str):
        key, progress, seft_grade = serial.split(":")
        self.key = key
        self.progress = int(progress)
        self.seft_grade = int(seft_grade)

    def __eq__(self, other):
        return self.progress == other.progress and self.seft_grade == other.seft_grade


    def __str__(self):
        return f'{self.key}:{self.progress}:{self.seft_grade}'
    

    
class DailyLog:
    def __init__(self, json_file: str):
        # dict[day, dict[task_key, TaskLog]]
        self.history: dict[str, dict[str, TaskData]] = {}
        self.json_file = json_file
        self.load_json()

        self.resume: dict[str, TaskData] = {}
        self.generate_resume()

    def load_json(self):
        history = {}
        try:
            if os.path.exists(self.json_file):
                with open(self.json_file, 'r') as f:
                    history = json.load(f)
        except:
            pass

        self.history = {}
        for day, serial_list in history.items():
            self.history[day] = {}
            for serial in serial_list:
                task_data = TaskData()
                task_data.decode(serial)
                self.history[day][task_data.key] = task_data

    def generate_resume(self):
        self.resume = {}
        #sort by tasklogrep by day
        for day in sorted(self.history.keys()):
            for task_key, task_log in self.history[day].items():
                self.resume[task_key] = task_log

    # def match_actual_state(self, day: str, actual_state: dict[str, TaskData]):
    #     if day not in self.history:
    #         self.history[day] = {}
    #     changed = False
    #     for _, task_data in actual_state.items():
    #         if self.__check_add_task(day, task_data):
    #                 changed = True
        
    #     if changed:
    #         self.save()

    def __check_add_task(self, day: str, task_data: TaskData) -> bool:
        data = self.resume.get(task_data.key, None)

        # não tinha, e ainda está zerado, não precisa salvar
        if data is None and task_data.progress == 0 and task_data.seft_grade == 0:
            return False
        
        # tinha, e está diferente, precisa salvar
        if data is None or data != task_data:
            if day not in self.history:
                self.history[day] = {}
            self.history[day][task_data.key] = task_data
            self.resume[task_data.key] = task_data
            return True
        return False
    
    def log_task(self, day: str, key: str, progress: int = -1, self_grade: int = -1, save_on_change: bool = True):
        actual_progress = self.resume.get(key, TaskData(key, 0, 0)).progress
        actual_seft_grade = self.resume.get(key, TaskData(key, 0, 0)).seft_grade
        if progress == -1:
            progress = actual_progress
        if self_grade == -1:
            self_grade = actual_seft_grade
        if self.__check_add_task(day, TaskData(key, progress, self_grade)):
            if save_on_change:
                self.save()
            return True
        return False

    def save(self):
        history = {}
        for day, task_data in self.history.items():
            history[day] = [str(task) for task in task_data.values()]
        with open(self.json_file, 'w') as f:
            json.dump(history, f, indent=4)



import datetime
import os
import shutil
import argparse
import csv

class Track:
    def __init__(self):
        self.timestamp = ""
        self.files: list[str] = []
        self.result: str = ""

    def set_timestamp(self, timestamp: str):
        self.timestamp = timestamp
        return self
    
    def set_files(self, files: list[str]):
        self.files = [os.path.basename(f) for f in files]
        return self
    
    def set_result(self, result: str):
        self.result = result
        return self
    
    def track_to_column(self):
        return [self.timestamp, self.result, ";".join(self.files)]


class Tracker:

    def __init__(self):
        self._files: list[str] = []
        self._folder: str = ""
        self._result: str = "None"
        self.log_file = "track.csv"
        self.track_folder = ".track"
        self.extension = ".log"
        self.posfix = ":"
        self.timestamp_size = len(self.get_timestamp() + self.posfix)
        self.extension_size = len(self.extension)

    def set_files(self, files: list[str]):
        self._files = [os.path.abspath(f) for f in files]
        return self
    
    def set_result(self, result: str):
        self._result = result
        return self
    
    def set_percentage(self, percentage: int):
        self._result = "{}%".format(str(percentage).ljust(3, "0"))
        return self

    def get_log_full_path(self):
        return os.path.join(self._folder, self.log_file)

    def set_folder(self, folder: str):
        self._folder = os.path.join(folder, self.track_folder)
        return self
    
    # in format: YYYY-MM-DD HH:MM:SS
    def get_timestamp(self):
        return datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    def get_file_history(self, file: str, file_list: list[str]):
        filename = os.path.basename(file)
        file_list = [f for f in file_list if f[self.timestamp_size:-self.extension_size] == filename]
        return sorted(file_list)

    def save_file_with_timestamp_prefix(self, timestamp: str, file: str) -> str:
        filename = os.path.basename(file)
        destination = os.path.join(self._folder, f"{timestamp}{self.posfix}{filename}{self.extension}")
        shutil.copy(file, destination)
        return destination
        

    def store(self):
        os.makedirs(self._folder, exist_ok=True)
        file_list = os.listdir(self._folder)
        file_list = [f for f in file_list if f.endswith(self.extension)]

        files_in_this_version: list[str] = []
        timestamp = self.get_timestamp()

        for file in self._files:
            saved_files = self.get_file_history(file, file_list)
            last_file = saved_files[-1] if len(saved_files) > 0 else ""
            if last_file == "":
                stored = self.save_file_with_timestamp_prefix(timestamp, file)
                files_in_this_version.append(stored)
                # print("Saved file: ", stored)
            else:
                change_time = os.path.getmtime(file)
                last_time = os.path.getmtime(os.path.join(self._folder, last_file))
                if change_time > last_time:
                    stored = self.save_file_with_timestamp_prefix(timestamp, file)
                    # print("Updated file: ", file)
                    files_in_this_version.append(stored)
                else:
                    files_in_this_version.append(last_file)
        

        log_file = self.get_log_full_path()
        track = Track().set_timestamp(timestamp).set_files(files_in_this_version).set_result(self._result)
        with open(log_file, mode="a") as f:
            writer = csv.writer(f)
            writer.writerow(track.track_to_column())

    @staticmethod
    def main():
        parser = argparse.ArgumentParser(description="Track files changes.")
        parser.add_argument("files", metavar="files", type=str, nargs="+", help="files to be tracked.")
        args = parser.parse_args()
        
        tracker = Tracker().set_folder(".track").set_files(args.files)
        tracker.store()



from __future__ import annotations
import csv
import hashlib
import datetime
from tko.settings.settings import Settings
from tko.util.daily import DailyLog
from tko.game.task import Task
import enum
import os

from abc import ABC, abstractmethod


class LogAction(enum.Enum):
    NONE = 'NONE'
    OPEN = 'OPEN'
    DOWN = 'DOWN'
    FREE = 'FREE' 
    TEST = 'TEST' # CORRECT|WRONG|COMPILE|RUNTIME
    SELF = 'SELF' # int
    PROG = 'PROG' # int
    QUIT = 'QUIT'


class ActionData:
    def __init__(self, timestamp: str, action_value: str, task: str = "", payload: str = ""):
        self.timestamp = timestamp
        self.action_value: str = action_value
        self.task_key = task
        self.payload = payload
        self.hash = ""

    def action_self_same_task(self, other: ActionData):
        if self.action_value != LogAction.SELF.value:
            return False
        if other.action_value != LogAction.SELF.value:
            return False
        return self.task_key == other.task_key

    def __str__(self):
        return f'{self.timestamp}, {self.action_value}, {self.task_key}, {self.payload}'

    @staticmethod
    def generate_hash(action_data: ActionData, last_hash: str):
        comprimento = 6
        input_str = str(action_data) + last_hash
        hash_completo = hashlib.sha256(input_str.encode()).hexdigest()
        return hash_completo[:comprimento]  # Retorna os primeiros 'comprimento' caracteres do hash

class LoggerStore(ABC):
    @abstractmethod
    def push_to_file(self, action_data: ActionData, last_hash: str):
        pass

    @abstractmethod
    def set_log_file(self, path: str):
        pass

    @abstractmethod
    def get_log_file(self) -> str | None:
        pass

    @abstractmethod
    def get_action_entries(self) -> list[ActionData]:
        pass

    @abstractmethod
    def get_last_hash(self) -> str:
        pass

class LoggerFS(LoggerStore):

    def __init__(self, settings: Settings):
        self.log_file: str | None = None
        self.settings = settings
    
    def set_log_file(self, log_file: str):
        self.log_file = log_file
        return self
    
    def get_log_file(self) -> str | None:
        return self.log_file
    
    def row_to_action_data(self, row: list[str]) -> ActionData:
        hash = row[0]
        timestamp = row[1]
        action_value = row[2]
        task = row[3]
        payload = row[4]
        action_data = ActionData(timestamp, action_value, task, payload)
        action_data.hash = hash
        return action_data

    def push_to_file(self, action_data: ActionData, last_hash: str):
        log_file = self.get_log_file()
        if log_file is None:
            return
        action_data.hash = ActionData.generate_hash(action_data, last_hash)
        if not os.path.exists(os.path.dirname(log_file)):
            os.makedirs(os.path.dirname(log_file))
        with open(log_file, 'a', newline='') as file:
            writer = csv.writer(file)
            ad = action_data
            writer.writerow([ad.hash, ad.timestamp, ad.action_value, ad.task_key, ad.payload])
            return ad.hash

    def get_action_entries(self) -> list[ActionData]:
        log_file = self.get_log_file()
        if log_file is None:
            return []
        if not os.path.exists(log_file):
            return []
        with open(log_file, 'r') as file:
            reader = csv.reader(file)
            rows = list(reader)
            return [self.row_to_action_data(row) for row in rows]

    def get_last_hash(self) -> str:
        return self.get_action_entries()[-1].hash if len(self.get_action_entries()) > 0 else ""

class LoggerMemory(LoggerStore):
    def __init__(self):
        self.entries: list[ActionData] = []

    def set_log_file(self, path: str):
        return self
    
    def get_log_file(self) -> str | None:
        return None

    def push_to_file(self, action_data: ActionData, last_hash: str):
        action_data.hash = ActionData.generate_hash(action_data, last_hash)
        self.entries.append(action_data)
        return action_data.hash

    def get_action_entries(self) -> list[ActionData]:
        return self.entries

    def get_last_hash(self) -> str:
        return self.entries[-1].hash if len(self.entries) > 0 else ""

class Logger:
    instance: None | Logger = None
    COMP_ERROR = "COMP"
    FREE_EXEC = "FREE"

    @staticmethod
    def get_instance() -> Logger:
        if Logger.instance is None:
            raise Exception("Logger not initialized")
        return Logger.instance

    def __init__(self, logger_store: LoggerStore):
        self.last_hash: str | None = None
        self.cached_action: ActionData | None = None
        self.fs = logger_store
        self.daily: DailyLog | None = None

    def set_history_file(self, log_file: str):
        self.fs.set_log_file(log_file)
        return self
    
    @staticmethod
    def today() -> str:
        return datetime.datetime.now().strftime('%Y-%m-%d')

    def set_daily(self, daily_file: str, tasks: dict[str, Task]):
        self.daily = DailyLog(daily_file)
        today = Logger.today()
        changed = False
        for task in tasks.values():
            if self.daily.log_task(today, task.key, task.progress, task.self_grade, save_on_change=False):
                changed = True
        if changed:
            self.daily.save()
        return self

    def check_log_file_integrity(self) -> list[str]:
        entries = self.fs.get_action_entries()
        if len(entries) == 0:
            return []
        hash = entries[0].hash
        output: list[str] = []

        for i in range(1, len(entries)):
            calculated_hash = ActionData.generate_hash(entries[i], hash)
            if calculated_hash != entries[i].hash:
                output.append(f"Hash mismatch line {i + 1}: {str(entries[i])}")
            hash = calculated_hash
        return output

    def store_in_cached(self, action_data: ActionData) -> bool:
        if self.cached_action is None and action_data.action_value == LogAction.SELF.value:
            return True
        if self.cached_action is not None and self.cached_action.action_self_same_task(action_data):
            return True
        return False

    def record_compilation_error(self, task_key: str):
        self.record_other_event(LogAction.TEST, task_key, self.COMP_ERROR)
    
    def record_test_result(self, task_key: str, result: int):
        self.record_other_event(LogAction.TEST, task_key, str(result))

    def record_freerun(self, task_key: str):
        self.record_other_event(LogAction.FREE, task_key)

    def record_other_event(self, action: LogAction, task_key: str = "", payload: str = ""):
        action_data = ActionData(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), action.value, task_key, payload)
        self.record_action_data(action_data)

    def get_last_hash(self) -> str:
        if self.last_hash is None:
            self.last_hash = self.fs.get_last_hash()
        return self.last_hash

    def record_action_data(self, action_data: ActionData):
        self.log_daily(action_data)

        if self.store_in_cached(action_data):
            self.cached_action = action_data
            return
        
        if self.cached_action is not None:
            self.last_hash = self.fs.push_to_file(self.cached_action, self.get_last_hash())
            self.cached_action = None
        
        if self.store_in_cached(action_data):
            self.cached_action = action_data
        else:
            self.last_hash = self.fs.push_to_file(action_data, self.get_last_hash())


    def log_daily(self, action_data: ActionData):
        if self.daily is None:
            return
        if action_data.action_value == LogAction.SELF.value:
            self.daily.log_task(Logger.today(), key=action_data.task_key, self_grade=int(action_data.payload))
        elif action_data.action_value == LogAction.PROG.value:
            self.daily.log_task(Logger.today(), key=action_data.task_key, progress=int(action_data.payload))
        elif action_data.action_value == LogAction.TEST.value:
            try:
                result = int(action_data.payload)                
                self.daily.log_task(Logger.today(), key=action_data.task_key, progress=result)
            except ValueError:
                pass


from tko.game.game import Game
from tko.game.graph import Graph
from tko.settings.repository import Repository
from tko.settings.settings import Settings
from tko.util.logger import Logger

import os

class CmdRep:
    @staticmethod
    def check(args):
        folder = Settings().get_alias_folder(args.alias)
        rep = Repository(folder).load_data_from_config_file().load_game()
        logger = Logger.get_instance()
        logger.set_history_file(rep.get_history_file())

        output = logger.check_log_file_integrity()
        if len(output) == 0:
            print(f"Arquivo de log do repositório {rep} está íntegro.")
        else:
            print(f"Arquivo de log do repositório {rep} está corrompido.")
            print("Erros:")
            for error in output:
                print(f"- {error}")

    @staticmethod
    def upgrade(args):
        folder = args.folder
        if os.path.exists(os.path.join(folder, "rep.json")):
            os.rename(os.path.join(folder, "rep.json"), os.path.join(folder, "repository.json"))
        remote_folder = os.path.join(folder, "remote")
        os.makedirs(remote_folder, exist_ok=True)
        for entry in os.listdir(folder):
            path = os.path.join(folder, entry)
            if entry == "remote":
                continue
            if os.path.isdir(path):
                os.rename(path, os.path.join(remote_folder, entry))
        print(f"Repositório {folder} foi atualizado.")

    @staticmethod
    def list(_args):
        settings = Settings()
        print(f"SettingsFile\n- {settings.settings_file}")
        print(str(settings))

    @staticmethod
    def add(args):
        settings = Settings().set_alias_remote(args.alias, args.value)
        settings.save_settings()

    @staticmethod
    def rm(args):
        sp = Settings()
        if args.alias in sp.dict_alias_remote:
            sp.dict_alias_remote.pop(args.alias)
            sp.save_settings()
        else:
            print("Repository not found.")

    @staticmethod
    def reset(_):
        sp = Settings().reset()
        print(sp.settings_file)
        sp.save_settings()

    @staticmethod
    def graph(args):
        settings = Settings()
        folder:str = settings.get_alias_folder(args.alias)
        rep = Repository(folder).load_data_from_config_file().load_game()
        rep.game.check_cycle()
        Graph(rep.game).generate()



from __future__ import annotations

import argparse
import sys

from tko.util.text import Text

from tko.cmds.cmd_rep import CmdRep
from tko.cmds.cmd_play import CmdPlay
from tko.cmds.cmd_down import CmdDown
from tko.cmds.cmd_run import Run
from tko.cmds.cmd_build import CmdBuild
from tko.cmds.cmd_config import CmdConfig, ConfigParams
from tko.util.logger import LogAction, Logger, LoggerFS
from tko.util.text import AnsiColor

from tko.util.param import Param
from tko.util.pattern import PatternLoader
from tko.util.consts import DiffCount, DiffMode
from tko.settings.settings import Settings

from tko.settings.repository import Repository

from tko.util.guide import tko_guide
from tko.util.guide import bash_guide

from tko.util.raw_terminal import RawTerminal
from tko.util.symbols import symbols
from tko.settings.check_version import CheckVersion

import os

from tko.__init__ import __version__


class Main:
    @staticmethod
    def test(args):
        PatternLoader.pattern = args.pattern
        param = Param.Basic().set_index(args.index)
        if args.quiet:
            param.set_diff_count(DiffCount.QUIET)
        elif args.all:
            param.set_diff_count(DiffCount.ALL)
        else:
            param.set_diff_count(DiffCount.FIRST)

        if args.filter:
            param.set_filter(True)
        if args.compact:
            param.set_compact(True)

        settings = Settings()
        if not args.side and not args.down:
            param.set_diff_mode(settings.app.get_diff_mode())
        elif args.side:
            param.set_diff_mode(DiffMode.SIDE)
        elif args.down:
            param.set_diff_mode(DiffMode.DOWN)
        cmd_run = Run(settings, args.target_list, args.cmd, param)
        cmd_run.execute()

    @staticmethod
    def run(args):
        PatternLoader.pattern = args.pattern
        param = Param.Basic().set_index(args.index)
        settings = Settings()
        param.set_diff_mode(settings.app.get_diff_mode())
        if args.filter:
            param.set_filter(True)
        cmd_run = Run(settings, args.target_list, args.cmd, param)
        cmd_run.set_curses()
        cmd_run.execute()

    @staticmethod
    def build(args):
        PatternLoader.pattern = args.pattern
        manip = Param.Manip().set_unlabel(args.unlabel).set_to_sort(args.sort).set_to_number(args.number)
        build = CmdBuild(args.target, args.target_list, manip)
        build.execute()

    @staticmethod
    def play(args):
        settings = Settings()
        if settings.has_alias_folder(args.alias):
            folder = settings.get_alias_folder(args.alias)
            CmdPlay(settings).load_folder(folder).execute()
        else:
            print(Text("{r}: não existe nenhuma repositório local cadastrado para o atalho {g}.", "Falha", args.alias))
        CheckVersion().version_check()

    @staticmethod
    def save(args):
        if not os.path.exists(args.folder) or not os.path.isdir(args.folder):
            print(Text("{r}: o diretório {g} não existe.", "Falha", args.folder))
            return

        settings = Settings()
        rep = Repository(args.folder)
        if not rep.has_local_config_file():
            print(Text("{r}: o diretório {g} não é um repositório válido.", "Falha", args.folder))
            return
        
        settings.set_alias_folder(args.alias, args.folder)
        print(Text("{g}: repositório {g} salvo com o atalho {g}.", "Sucesso", args.folder, args.alias))

    # @staticmethod
    # def play(args):
    #     print(Text("{g}: utilize o comando '{y}' para abrir um repositório local cadastrado.", "Atenção", "tko load " + args.alias))

    @staticmethod
    def open(args):
        settings = Settings()
        CmdPlay(settings).load_folder(args.folder).execute()

    @staticmethod
    def init(args):
        remote: str | None = args.remote
        url: str | None = args.url
        file: str | None = args.file
        folder: str | None = args.folder
        save: str | None = args.save
        Main.__init(remote=remote, url=url, file=file, folder=folder, save=save)

    @staticmethod
    def __init(remote: str | None, url: str | None, file: str | None, folder: str | None, save: str | None):
        if folder is None:
            folder = os.getcwd()
            print("Deseja criar um repositório em " + folder + "? (s/n): ", end="")
            op = input()
            if op == "n":
                return
        rep = Repository(os.path.abspath(folder))
        if remote is None and url is None and file is None:
            index = rep.get_default_readme_path()
            print("Nenhuma fonte foi informada, utilizando o arquivo {} como fonte".format(index))
            if not os.path.exists(index):
                with open(index, "w") as f:
                    f.write("# Repositório\n\n")
                    f.write("## Grupo\n\n")
                    f.write("### Missão\n\n")
                    f.write("- [ ] [#google Abra o google](https://www.google.com)\n")
        else:
            settings = Settings()
            source: str = ""
            if remote is not None:
                if settings.has_alias_remote(remote):
                    source = settings.get_alias_remote(remote)
                else:
                    raise Warning("fail: alias remoto não encontrado.")
            elif url is not None:
                source = url
            elif file is not None:
                source = file
            rep.set_remote_source(source)
        rep.save_data_to_config_file()
        if save is not None:
            settings = Settings()
            settings.set_alias_folder(save, folder)
            settings.save_settings()

    @staticmethod
    def start(args):
        remote: str = args.remote
        Main.__init(remote=remote, url=None, file=None, folder=remote, save=remote)
        print(Text("Repositório {g} criado com sucesso na pasta {g}", remote, os.path.abspath(remote)))

    # @staticmethod
    # def down(args):
    #     settings = Settings()
    #     settings.check_rootdir()
    #     cmd_down = CmdDown(rep=args.course, task_key=args.activity, settings=settings)
    #     cmd_down.set_language(args.language).set_fnprint(print)
    #     cmd_down.execute()

    @staticmethod
    def config(args):
        settings = Settings()
        param = ConfigParams()
        param.side = args.side
        param.down = args.down
        param.images = args.images
        param.editor = args.editor
        param.borders = args.borders
        param.timeout = args.timeout

        CmdConfig.execute(settings, param)

    @staticmethod
    def list(_args):
        settings = Settings()
        print(str(settings))


class Parser:
    def __init__(self):
        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(prog='tko', description=f'tko version {__version__}')
        self.subparsers = self.parser.add_subparsers(title='subcommands', help='help for subcommand.')

        self.parent_manip = self.create_parent_manip()
        self.parent_basic = self.create_parent_basic()

        self.add_parser_global()
        self.create_parent_basic()
        self.create_parent_manip()
        self.add_parser_run()
        self.add_parser_exec()
        self.add_parser_build()
        # self.add_parser_down()
        self.add_parser_config()
        self.add_parser_rep_tools()
        self.add_parser_rep_actions()
        # self.add_parser_open()
        # self.add_parser_save()
        # self.add_parser_init()


    def add_parser_global(self):
        self.parser.add_argument('-c', metavar='CONFIG_FILE', type=str, help='config json file.')
        self.parser.add_argument('-w', metavar='WIDTH', type=int, help="terminal width.")
        self.parser.add_argument('-v', action='store_true', help='show version.')
        self.parser.add_argument('-m', action='store_true', help='monochromatic.')

    def create_parent_basic(self):
        parent_basic = argparse.ArgumentParser(add_help=False)
        parent_basic.add_argument('--index', '-i', metavar="I", type=int, help='run a specific index.')
        parent_basic.add_argument('--pattern', '-p', metavar="P", type=str, default='@.in @.sol',
                                  help='pattern load/save a folder, default: "@.in @.sol"')
        return parent_basic

    def create_parent_manip(self):
        parent_manip = argparse.ArgumentParser(add_help=False)
        parent_manip.add_argument('--width', '-w', type=int, help="term width.")
        parent_manip.add_argument('--unlabel', '-u', action='store_true', help='remove all labels.')
        parent_manip.add_argument('--number', '-n', action='store_true', help='number labels.')
        parent_manip.add_argument('--sort', '-s', action='store_true', help="sort test cases by input size.")
        parent_manip.add_argument('--pattern', '-p', metavar="@.in @.out", type=str, default='@.in @.sol',
                                  help='pattern load/save a folder, default: "@.in @.sol"')
        return parent_manip

    def add_parser_run(self):
        parser_r = self.subparsers.add_parser('run', parents=[self.parent_basic], help='run with test cases using curses.')
        parser_r.add_argument('target_list', metavar='T', type=str, nargs='*', help='solvers, test cases or folders.')
        parser_r.add_argument('--filter', '-f', action='store_true', help='filter solver in temp dir before run')
        parser_r.add_argument("--cmd", type=str, help="bash command to run code")
        parser_r.set_defaults(func=Main.run)

    def add_parser_exec(self):
        parser_r = self.subparsers.add_parser('test', parents=[self.parent_basic], help='run with test cases.')
        parser_r.add_argument('target_list', metavar='T', type=str, nargs='*', help='solvers, test cases or folders.')
        parser_r.add_argument('--filter', '-f', action='store_true', help='filter solver in temp dir before run')
        parser_r.add_argument('--compact', '-c', action='store_true', help='Do not show case descriptions in failures')
        parser_r.add_argument("--cmd", type=str, help="bash command to run code")

        group_n = parser_r.add_mutually_exclusive_group()
        group_n.add_argument('--quiet', '-q', action='store_true', help='quiet mode, do not show any failure.')
        group_n.add_argument('--all', '-a', action='store_true', help='show all failures.')

        # add an exclusive group for diff mode
        group = parser_r.add_mutually_exclusive_group()
        group.add_argument('--down', '-d', action='store_true', help="diff mode up-to-down.")
        group.add_argument('--side', '-s', action='store_true', help="diff mode side-by-side.")
        parser_r.set_defaults(func=Main.test)

    def add_parser_build(self):
        parser_b = self.subparsers.add_parser('build', parents=[self.parent_manip], help='build a test target.')
        parser_b.add_argument('target', metavar='T_OUT', type=str, help='target to be build.')
        parser_b.add_argument('target_list', metavar='T', type=str, nargs='+', help='input test targets.')
        parser_b.set_defaults(func=Main.build)

    # def add_parser_down(self):
    #     parser_d = self.subparsers.add_parser('down', help='download problem from repository.')
    #     parser_d.add_argument('course', type=str, nargs='?', help=" [ fup | ed | poo ].")
    #     parser_d.add_argument('activity', type=str, nargs='?', help="activity @label.")
    #     parser_d.add_argument('--language', '-l', type=str, nargs='?', help="[ c | cpp | js | ts | py | java ]")
    #     parser_d.set_defaults(func=Main.down)

    def add_parser_config(self):
        parser_cfg = self.subparsers.add_parser('config', help='settings tool.')
        subpar_repo = parser_cfg.add_subparsers(title='subcommands', help='help for subcommand.')

        cfg_reset = subpar_repo.add_parser('reset', help='reset all repositories, folders and values to factory default.')
        cfg_reset.set_defaults(func=CmdRep.reset)

        cfg_set = subpar_repo.add_parser('set', help='set default config values.')

        g_diff = cfg_set.add_mutually_exclusive_group()
        g_diff.add_argument('--side', action='store_true', help='set side_by_side diff mode.')
        g_diff.add_argument('--down', action='store_true', help='set up_to_down   diff mode.')

        # g_lang = parser_s.add_mutually_exclusive_group()
        # g_lang.add_argument("--lang", '-l', metavar='ext', type=str, help="set default language extension.")
        # g_lang.add_argument("--ask", action='store_true', help='ask language extension every time.')

        # parser_s.add_argument("--root", metavar="path", type=str, help='set root directory.')
        cfg_set.add_argument("--editor", metavar="cmd", type=str, help='set editor command.')
        cfg_set.add_argument("--borders", metavar="[0|1]", type=str, help='enable borders.')
        cfg_set.add_argument("--images", metavar="[0|1]", type=str, help='enable images.')
        cfg_set.add_argument("--timeout", metavar="sec", type=int, help='set timeout.')

        cfg_set.set_defaults(func=Main.config)

        cfg_list = subpar_repo.add_parser('list', help='list default config values.')
        cfg_list.set_defaults(func=Main.list)



    def add_parser_rep_tools(self):

        parser_repo = self.subparsers.add_parser('rep', help='Repository tools.')
        subpar_repo = parser_repo.add_subparsers(title='subcommands', help='help for subcommand.')

        # repo_reset = subpar_repo.add_parser('reset', help='reset all repositories and folders to factory default.')
        # repo_reset.set_defaults(func=CmdRep.reset)

        # repo_graph = subpar_repo.add_parser('graph', help='generate graph of the repository.')
        # repo_graph.add_argument('alias', metavar='alias', type=str, help='alias of the repository to be graphed.')
        # repo_graph.set_defaults(func=CmdRep.graph)

        repo_log = subpar_repo.add_parser("check", help="validates log of the repository.")
        repo_log.add_argument('folder', type=str, help='folder to be checked.')
        repo_log.set_defaults(func=CmdRep.check)

    # def add_parser_save(self):
    #     parser_p = self.subparsers.add_parser('save', help='Save a repository folder with a global alias.')
    #     parser_p.add_argument('folder', type=str, help='repository folder.')
    #     parser_p.add_argument('alias', type=str, help='repository alias.')
    #     parser_p.set_defaults(func=CmdRep.save)

    def add_parser_rep_actions(self):
        # parser_load = self.subparsers.add_parser('load', help='Load a saved repository.')
        # parser_load.add_argument('alias', type=str, help='repository alias.')
        # parser_load.set_defaults(func=Main.load)

        parser_play = self.subparsers.add_parser('play', help='Load and play a saved repository.')
        parser_play.add_argument('alias', type=str, help='repository alias.')
        parser_play.set_defaults(func=Main.play)

        parser_open = self.subparsers.add_parser('open', help='Open a folder with a repository.')
        parser_open.add_argument('folder', type=str, help='folder.')
        parser_open.set_defaults(func=Main.open)

        parser_save = self.subparsers.add_parser('save', help='Save a repository in a global alias.')
        parser_save.add_argument('alias', type=str, help='alias.')
        parser_save.add_argument('folder', type=str, help='folder.')
        parser_save.set_defaults(func=Main.save)

        parser_init = self.subparsers.add_parser('init', help='Initialize a repository in a folder.')

        source = parser_init.add_mutually_exclusive_group()
        source.add_argument('--remote', type=str, help='remote source [fup|ed|poo].')
        source.add_argument('--url', type=str, help='remote url.')
        source.add_argument('--file', type=str, help='remote or local file.')

        parser_init.add_argument('--folder', type=str, help='local directory.')
        parser_init.add_argument('--save', metavar="alias", type=str, help='save in global alias.')
        parser_init.set_defaults(func=Main.init)

        parser_start = self.subparsers.add_parser('start', help='Initialize a repository in a folder with a remote source and save in alias.')
        parser_start.add_argument('remote', type=str, help='remote source [fup|ed|poo].')
        parser_start.set_defaults(func=Main.start)


    # def add_parser_open(self):


    # def add_parser_init(self):
    #     parser_p = self.subparsers.add_parser('open', help='open a folder with a repository.')
    #     parser_p.add_argument('folder', metavar='folder', type=str, help='folder.')
    #     parser_p.set_defaults(func=CmdRep.init)



def exec(parser: argparse.ArgumentParser, args):
    settings = Settings()
    if args.w is not None:
        RawTerminal.set_terminal_size(args.w)
    if args.c:
        settings.set_settings_file(args.c)
    settings.load_settings()
    Logger.instance = Logger(LoggerFS(settings))

    if args.m:
        AnsiColor.enabled = False
    else:
        AnsiColor.enabled = True
        symbols.set_colors()

    if args.v:
        if args.v:
            print("tko version " + __version__)
    else:
        if "func" in args:
            args.func(args)
        else:
            parser.print_help()

def main():
    try:
        parser = Parser().parser
        args = parser.parse_args()
        exec(parser, args)
        sys.exit(0)
    except KeyboardInterrupt:
        print("\n\nKeyboard Interrupt")
        sys.exit(1)
    except Warning as w:
        print(w)
        sys.exit(1)
    # except Exception as e:
    #     print(e)
    #     sys.exit(1)

# NÃO RETIRE ESSA TAG POR CAUSA DO MERGE DO REPLIT
if __name__ == '__main__': # MERGE_INSERT 
    main()



from typing import Dict, Optional, Tuple
from tko.util.symbols import symbols
from tko.util.text import Text
# from tko.util.logger import Logger, LogAction
from tko.game.tree_item import TreeItem
import re
import os

class Task(TreeItem):

    def __init__(self):
        super().__init__()
        self.line_number = 0
        self.line = ""
        self.downloadable = False
        self.self_grade: int = 0 #valor de 0 a 9
        self.progress: int = 0 #valor de 0 a 100
        self.main_index: int = 0

        self.qskills: Dict[str, int] = {} # default quest skills
        self.skills: Dict[str, int] = {} # local skills
        self.xp: int = 0
        
        self.opt: bool = False
        self.link: str = ""

        self.quest_key = ""
        self.cluster_key = ""
        self.folder: str = ""

        self.default_min_value = 7 # default min grade to complete task

    def set_folder(self, folder: str):
        self.folder = folder
        return self

    def load_from_db(self, value: str):
        if ":" not in value:
            self.self_grade = int(value)
        else:
            v = value.split(":")
            if len(v) == 3:
                self.self_grade = int(v[0])
                self.main_index = int(v[1])
                self.progress = int(v[2])

    def save_to_db(self) -> str:
        return f"{self.self_grade}:{self.main_index}:{self.progress}"
    
    def is_db_empty(self) -> bool:
        return self.self_grade == 0 and self.main_index == 0 and self.progress == 0

    def get_prog_color(self, min_value: Optional[int] = None) -> str:
        if min_value is None:
            min_value = self.default_min_value
        prog = self.progress // 10
        if prog == 0:
            return "m"
        if prog < min_value:
            return "r"
        if prog < 10:
            return "y"
        if prog == 10:
            return "g"
        return "w"  

    def get_prog_symbol(self, min_value: Optional[int] = None) -> Text:
        
        if min_value is None:
            min_value = self.default_min_value
        color = self.get_prog_color(min_value)
        prog = self.progress // 10
        if prog == 0:
            return Text().addf(color, symbols.uncheck.text)
        if prog < min_value:
            return Text().addf(color, str(prog))
        if prog < 10:
            return Text().addf(color, str(prog))
        if prog == 10:
            return Text().addf(color, symbols.check.text)
        return Text().add("0")

    def get_percent(self):
        if self.self_grade == 0:
            return 0
        if self.self_grade == 10:
            return 100
        return self.self_grade * 10
    
    def is_complete(self):
        return self.self_grade == 10

    def not_started(self):
        return self.self_grade == 0
    
    def in_progress(self):
        return self.self_grade > 0 and self.self_grade < 10

    def set_grade(self, grade: int):
        grade = int(grade)
        if grade >= 0 and grade <= 10:
            if grade != self.self_grade:
                self.self_grade = grade
        else:
            print(f"Grade inválida: {grade}")
    
    def process_link(self, base_file):
        if self.link.startswith("http"):
            return
        if self.link.startswith("./"):
            self.link = self.link[2:]
        # todo trocar / por \\ se windows
        self.link = base_file + self.link

    def __str__(self):
        line = str(self.line_number).rjust(3)
        key = "" if self.key == self.title else self.key + " "
        return f"{line}    {self.self_grade} {key}{self.title} {self.skills} {self.link}"
    
    def has_remote_link(self):
        return self.link.startswith("http:") or self.link.startswith("https:")

    def has_at_symbol(self):
        return any([s.startswith("@") for s in self.title.split(" ")])

    def is_downloadable(self):
        return self.downloadable

    def is_downloaded_for_lang(self, task_dir: str, lang: str) -> bool:
        if not os.path.isfile(os.path.join(task_dir, "Readme.md")):
            return False
        files = os.listdir(task_dir)
        if not any([f.endswith("." + lang) for f in files]):
            return False
        return True

"""
Usar @ ou # para definir a chave da tarefa
Pode ser incluído no título da tarefa que está entre os [] ou nas tags url
Exemplo:
- [ ] [@banana Título da tarefa](https://tal_coias/Readme.md) 
Chave: "banana"
Título: "@banana Título da tarefa"
link: "https://tal_coias/Readme.md"
É uma tarefa de download

----------
Nessas tarefas que não serão baixadas para o repositório, se não for inserido o #, será utilizado como chave o link da tarefa
- [ ] [#banana Título da tarefa](https://tal_coias/Readme.md)
Chave: "banana"
Título: "#banana Título da tarefa"
link: "https://tal_coias/Readme.md"


"""
class TaskParser:

    @staticmethod
    def __load_tags(task: Task, tags_raw: str):
        tags = [tag.strip() for tag in tags_raw.split(" ")]
        task.opt = "opt" in tags
        for t in tags:
            if t.startswith("+"):
                key, value = t[1:].split(":")
                task.skills[key] = int(value)
            elif t.startswith("@"):
                task.key = t[1:]
                task.downloadable = True
            elif t.startswith("#"):
                task.key = t[1:]

    @staticmethod
    def parse_line(line: str, line_num: int) -> Optional[Task]:
        pattern = r'\s*?- \[ \](.*?)\[([^\]]+)\]\(([^)]+)\)(?:\s*<!--(.*?)-->)?'

        match = re.match(pattern, line)
        if match is None:
            return None
        task = Task()
        task.line_number = line_num
        task.line = line
        task.opt = False
        task.title = match.group(1).strip()
        if task.title != "":
            task.title += " "
        task.title += match.group(2).strip()
        task.title = task.title.replace("`", "")
        task.link = match.group(3).strip()
        if match.group(4) is not None:
            TaskParser.__load_tags(task, match.group(3))
        
        for item in task.title.split(" "):
            if item.startswith("@"):
                task.key = item[1:]
                task.downloadable = True
            if item.startswith("#"):
                task.key = item[1:]
        if task.key == "":
            task.key = task.link
        return task


import json
from typing import Any, Dict
from .repository import Repository
from .app_settings import AppSettings
import os
import appdirs

from tko.util.text import Text
from tko.play.colors import Colors

def singleton(class_):
    instances = {}
    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]
    return getinstance

@singleton
class Settings:
    CFG_FILE = "settings.json"


    def __init__(self):
        self.__remote = "remote"
        self.__folder = "folders"
        self.__appcfg = "appcfg"
        self.__colors = "colors"

        self.dict_alias_remote: Dict[str, str] = {}
        self.dict_alias_folder: Dict[str, str] = {}
        self.app = AppSettings()
        self.colors = Colors()

        self.settings_file = ""

    def set_settings_file(self, path: str):
        self.settings_file = path
        return self

    def get_settings_file(self) -> str:
        if self.settings_file is None or self.settings_file == "":
            self.package_name = "tko"
            default_filename = self.CFG_FILE
            self.settings_file = os.path.abspath(default_filename)  # backup for replit, dont remove
            self.settings_file = os.path.join(appdirs.user_data_dir(self.package_name), default_filename)
        
        if not os.path.exists(self.settings_file):
            os.makedirs(os.path.dirname(self.settings_file), exist_ok=True)
        return self.settings_file

    def reset(self):
        self.dict_alias_remote = {}
        self.dict_alias_remote["fup"] = "https://github.com/qxcodefup/arcade/blob/master/Readme.md"
        self.dict_alias_remote["ed"] = "https://github.com/qxcodeed/arcade/blob/master/Readme.md"
        self.dict_alias_remote["poo"] = "https://github.com/qxcodepoo/arcade/blob/master/Readme.md"
        self.dict_alias_folder = {}
        self.app = AppSettings()
        self.colors = Colors()
        return self

    def set_alias_remote(self, alias: str, url_or_path: str):
        if not(url_or_path.startswith("http:") or url_or_path.startswith("https:")):
            url_or_path = os.path.abspath(url_or_path)
        self.dict_alias_remote[alias] = url_or_path
        return self

    def has_alias_remote(self, alias: str) -> bool:
        return alias in self.dict_alias_remote

    def get_alias_remote(self, alias: str) -> str:
        if alias in self.dict_alias_remote:
            return self.dict_alias_remote[alias]
        raise Warning(f"Repositório remoto {alias} não encontrado")

    def set_alias_folder(self, alias: str, folder: str):
        self.dict_alias_folder[alias] = os.path.abspath(folder)
        return self
    
    def has_alias_folder(self, course: str) -> bool:
        return course in self.dict_alias_folder

    def del_alias_folder(self, course: str):
        if course in self.dict_alias_folder:
            del self.dict_alias_folder[course]
        return self

    def get_alias_folder(self, course: str) -> str:
        if course in self.dict_alias_folder:
            folder = self.dict_alias_folder[course]
            if not isinstance(folder, str):
                self.reset()
                self.save_settings()
                return self.get_alias_folder(course)
            else:
                return folder
        raise Warning(f"Não registrada pasta local para {course}")
    
    def load_settings(self):
        try:
            settings_file = self.get_settings_file() # assure right loading if value == ""
            with open(settings_file, "r", encoding="utf-8") as f:
                data = json.load(f)
                self.dict_alias_remote = data.get(self.__remote, {})
                self.dict_alias_folder = data.get(self.__folder, {})
                self.app = AppSettings().from_dict(data.get(self.__appcfg, {}))
                self.colors = Colors().from_dict(data.get(self.__colors, {}))
        except (FileNotFoundError, json.decoder.JSONDecodeError) as _e:
            self.reset()
            self.save_settings()
        return self

    # def check_rootdir(self):
    #     if self.app.get_rootdir() != "":
    #         return self
    #     print(Text().add("Pasta padrão para download de arquivos ").addf("r", "precisa").add(" ser definida."))
    #     here_cwd = os.getcwd()
    #     qxcode = os.path.join(os.path.expanduser("~"), "qxcode")

    #     while True:
    #         print(Text().addf("r", "1").add(" - ").add(here_cwd))
    #         print(Text().addf("r", "2").add(" - ").add(qxcode))
    #         print(Text().addf("r", "3").add(" - ").add("Outra pasta"))
    #         print(Text().add("Default ").addf("r", "1").add(": "), end="")
    #         op = input()
    #         if op == "":
    #             op = "1"
    #         if op == "1":
    #             home_qxcode = here_cwd
    #             break
    #         if op == "2":
    #             home_qxcode = qxcode
    #             break
    #         if op == "3":
    #             print(Text().addf("y", "Navegue até o diretório desejado e execute o tko novamente."))
    #             exit(1)

    #     if not os.path.exists(home_qxcode):
    #         os.makedirs(home_qxcode)
    #     print("Pasta padrão para download de arquivos foi definida em: " + home_qxcode)
    #     print(RawTerminal.centralize("", "-"))
    #     self.app._rootdir = home_qxcode
    #     self.save_settings();
    #     return self
    
    # def check_rep_alias(self, rep_alias: str):
    #     if rep_alias == "__ask":
    #         last = self.app.get_last_rep()
    #         if last != "" and last in self.remote:
    #             rep_alias = last
    #         else:
    #             print("Escolha um dos repositórios para abrir:")
    #             options: Dict[int, str] = {}
    #             for i, alias in enumerate(self.remote, start=1):
    #                 print(Text().addf("r", str(i)).add(f" - {alias}"))
    #                 options[i] = alias
    #             while True:
    #                 try:
    #                     print("Digite o número do repositório desejado: ", end="")
    #                     index = int(input())
    #                     if index in options:
    #                         rep_alias = options[index]
    #                         self.app.set_last_rep(rep_alias)
    #                         self.save_settings()
    #                         break
    #                 except ValueError:
    #                     pass
    #                 print("Digite um número válido")
    #     return rep_alias
    
    def save_settings(self):
        file = self.get_settings_file()
        value = {
            self.__remote: self.dict_alias_remote,
            self.__folder: self.dict_alias_folder,
            self.__appcfg: self.app.to_dict(),
            self.__colors: self.colors.to_dict()
        }
        with open(file, "w", encoding="utf-8") as f:
            json.dump(value, f, indent=4)
        return self

    def __str__(self):
        output = []
        output.append(str(Text("{g}", "Arquivo de configuração:")))
        output.append("- " + self.get_settings_file())
        output.append("")
        output.append(str(Text("{g}", "Repositórios remotos cadastrados:")))
        max_alias = max([len(key) for key in self.dict_alias_remote])
        for key in self.dict_alias_remote:
            output.append("- {} : {}".format(key.ljust(max_alias), self.dict_alias_remote[key]))
        output.append("")
        output.append(str(Text("{g}", "Repositórios locais salvos:")))
        max_alias = max([len(key) for key in self.dict_alias_folder])
        for key in self.dict_alias_folder:
            output.append("- {} : {}".format(key.ljust(max_alias), self.dict_alias_folder[key]))

        app_str = str(self.app)
        return "\n".join(output) + "\n\n" + app_str

        # maxlen = max([len(key) for key in self.remote])
        # for key in self.remote:
        #     prefix = f"- {key.ljust(maxlen)}"
        #     if self.remote[key].file and self.remote[key].url:
        #         output.append(f"{prefix} : dual   : {self.remote[key].url} ; {self.remote[key].file}")
        #     elif self.remote[key].url:
        #         output.append(f"{prefix} : remote : {self.remote[key].url}")
        #     else:
        #         output.append(f"{prefix} : local  : {self.remote[key].file}")
        # return "\n".join(output)



from typing import Optional, List, Callable
import os
import urllib.request
import urllib.error
import json

from tko.settings.repository import Repository
from tko.down.down import DownProblem

from tko.settings.settings import Settings
from tko.game.game import Game
from tko.util.remote_url import RemoteUrl
from tko.cmds.cmd_build import CmdBuild
from tko.util.param import Param


class CmdDown:
    def __init__(self, rep: Repository, task_key: str, settings: Settings):
        self.rep = rep
        self.task_key = task_key
        self.settings = settings
        self.game: Game | None = None
        self.language: str | None = None
        self.destiny_folder: str = ""
        self.readme_path: str = ""
        self.cache_url: str = ""
        self.mapi_file: str = ""

    def set_fnprint(self, fnprint: Callable[[str], None]):
        DownProblem.fnprint = fnprint
        return self

    def set_game(self, game: Game):
        self.game = game
        return self
    
    def set_language(self, language: str):
        self.language = language
        return self
    
    def get_game(self):
        if self.game is not None:
            return self.game
        try:
            file = self.rep.load_index_or_cache()
        except urllib.error.HTTPError:
            DownProblem.fnprint("falha: Verifique sua internet")
        return Game(file)

    def remove_empty_destiny_folder(self):
        if len(os.listdir(self.destiny_folder)) == 0:
                os.rmdir(self.destiny_folder)

    def download_readme(self, readme_remote_url) -> bool:
        DownProblem.create_problem_folder(self.destiny_folder)
        try:
            DownProblem.down_readme(self.readme_path, readme_remote_url)
            return True
        except urllib.error.HTTPError:
            DownProblem.fnprint("falha: Arquivo de descrição não encontrado")
            return False
        except urllib.error.URLError:
            DownProblem.fnprint("falha: Não consegui baixar a atividade, verifique sua internet")
            return False

    def download_mapi(self) -> bool:
        try:
            urllib.request.urlretrieve(self.cache_url + "mapi.json", self.mapi_file)
            return True
        except urllib.error.HTTPError:
            return False
        except urllib.error.URLError:
            DownProblem.fnprint("falha: não consegui baixar a atividade, verifique sua internet")
            return False

    def build_cases_from_readme(self, destiny_folder: str):
        cases_tio_target = os.path.join(destiny_folder, "cases.tio")
        param = Param.Manip()
        cb = CmdBuild(cases_tio_target, [self.readme_path], param)
        cb.set_quiet(True)
        cb.execute()

    def execute(self) -> bool:
        item = self.get_game().get_task(self.task_key)
        if not item.link.startswith("http"):
            DownProblem.fnprint("falha: link para atividade não é um link remoto")
            return False
        
        readme_remote_url = RemoteUrl(item.link)
        self.cache_url = os.path.dirname(readme_remote_url.get_raw_url()) + "/.cache/"
        self.destiny_folder = self.rep.get_remote_task_folder(self.task_key)
        self.readme_path =  os.path.join(self.destiny_folder, "Readme.md")
        self.mapi_file = os.path.join(self.destiny_folder, "mapi.json")
        if not self.download_readme(readme_remote_url):
            return False
        
        lang = self.check_and_select_language()
        
        if self.download_mapi():
            with open(self.mapi_file, encoding="utf-8") as f:
                loaded_json = json.load(f)
            os.remove(self.mapi_file)
            DownProblem.unpack_json(loaded_json, self.destiny_folder, lang)
            if not DownProblem.check_draft_existence(loaded_json, self.destiny_folder, lang, self.cache_url):
                DownProblem.create_default_draft(self.destiny_folder, lang)
        else:
            self.build_cases_from_readme(self.destiny_folder)
            DownProblem.create_default_draft(self.destiny_folder, lang)
        DownProblem.fnprint("")
        DownProblem.fnprint("Atividade baixada com sucesso")
        return True

    def check_and_select_language(self) -> str:
        language_def = self.rep.get_lang()

        if self.language is None:
            if language_def != "":
                self.language = language_def
            else:
                print("Escolha uma extensão para os rascunhos: [c, cpp, py, ts, js, java]: ", end="")
                self.language = input()
        return self.language
